"""
module for parsing doctrings to produce a .json file for swagger-ui
"""

from spynl.main.docs.documentation import (document_endpoint,
                                           add_swagger_reusables)



"""Handle YAML content"""

import re

import yaml

from spynl.main.serial.exceptions import MalformedRequestException


expression = re.compile(r'^\s*\-')


def sniff(body):
    """Sniff body content, return True if YAML detected"""
    return bool(re.match(expression, body))


def dumps(body, pretty=False):
    """return YAML body as string"""
    if pretty:
        return yaml.dump(body, indent=4)
    else:
        return yaml.dump(body)


def loads(body, headers=None):
    """return body as YAML"""
    try:
        return yaml.load(body)
    except ValueError as e:
        raise MalformedRequestException('application/x-yaml', str(e))



"""
Some very basic endpoints that do not require permission:
    * ping
    * time
    * error
    * error500

Also, two endpoints to test requests:
    * request_echo
    * request_check
"""

from datetime import datetime

from pyramid.security import NO_PERMISSION_REQUIRED

from spynl.main.utils import get_user_info
from spynl.main.dateutils import now, localize_date, date_to_str


def ping(request):
    """
    Ping Spynl.

    ---
    get:
      tags:
        - contact
      description: >
        Spynl tells you "pong" and the time.

        ####Response

        JSON keys | Content Type | Description\n
        --------- | ------------ | -----------\n
        status    | string | 'ok' or 'error'\n
        time      | string | time in format: $setting[spynl.date_format]\n
        greeting  | string | 'pong'\n
    """
    return {'time': now(), 'greeting': 'pong'}


def time(request):
    """
    Get the Spynl time.

    ---
    get:
      tags:
        - contact
      description: >
        Spynl tells you its server time and what it believes to be
        your local time if you are logged in.

        ####Response

        JSON keys | Content Type | Description\n
        --------- | ------------ | -----------\n
        status    | string | 'ok' or 'error'\n
        server_time | string | server time in format:
        $setting[spynl.date_format]\n
        local_time  | string | local user time in format:
        $setting[spynl.date_format]\n
        tz        | string | local time zone, e.g. Europe/Amsterdam\n
    """
    response = {'server_time': date_to_str(localize_date(datetime.utcnow(),
                                                         user_specific=False))}
    user_info = get_user_info(request)
    if user_info.get('tz') is not None:
        response['local_time'] = date_to_str(now())
        response['tz'] = user_info.get('tz', None)

    return response


def request_echo(request):
    """return request args (see utils.get_args) - for testing"""
    return request.args


def request_check(r):
    """return a valid json response with meta data from request"""
    request = {'status': 'ok'}
    if r.args.get('data'):
        request['data'] = r.args.get('data')
    if r.args.get('method'):
        request['method'] = r.args.get('method')
    if r.args.get('resource'):
        request['resource'] = r.args.get('resource')
    return request


def main(config):
    """
    Add two basic views, two error views and two test views
    """
    config.add_endpoint(ping, 'ping', permission=NO_PERMISSION_REQUIRED)
    config.add_endpoint(time, 'time', permission=NO_PERMISSION_REQUIRED)

    # useful for testing purposes
    config.add_endpoint(request_echo, 'request_echo',
                        permission=NO_PERMISSION_REQUIRED)
    config.add_endpoint(request_check, 'request_check',
                        permission=NO_PERMISSION_REQUIRED)



"""If needed, Spynl can return an HTML response."""

from spynl.main.serial.xml import prettify_xml


DOCTYPE = '<!DOCTYPE html>'


def dumps(content, pretty=False):
    """format an HTML response"""
    # only prettify HTML if necessary
    if pretty and "<html><" in content and "<body><" in content:
        tags = ['<' + e for e in content.split('<') if e != ""]
        content = ''.join(prettify_xml(tags))
    if not content.lstrip().startswith("<!DOCTYPE"):
        content = DOCTYPE + content
    return '{}{}'.format(pretty * '\n', content)



"""function to dump python objects"""

from pprint import pprint
from io import StringIO


def dumps(body, pretty=False):
    """Return Python objects in body in text representation"""
    if pretty:
        stream = StringIO()
        pprint(body, stream=stream)
        return stream.getvalue()

    return repr(body)



"""
support checking Spynl request or response data against schemas
"""

import os
import json
from functools import wraps

import jsonschema

from spynl.main.utils import get_logger, get_yaml_from_docstring
from spynl.main.exceptions import InvalidResponse, BadValidationInstructions
from spynl.main.locale import SpynlTranslationString as _


def validate_json_schema(endpoint, info):
    """
    Validate JSON in- and output against a schema
    if the endpoint definition says so.

    This is a view deriver for Spynl endpoints.
    """
    if info.options.get('is_error_view', False) is True:
        return endpoint  # no need to validate (again)
    @wraps(endpoint)
    def validate(context, request):
        """Function wrapping our endpoint."""
        yaml_doc = get_yaml_from_docstring(endpoint.__doc__)
        validations = []
        if yaml_doc and 'validations' in yaml_doc:
            validations = yaml_doc['validations']
        # try to validate request body
        if request.content_type == 'application/json'\
            and hasattr(request, 'args'):
            interprete_validation_instructions(request.args, validations,
                                               'request')
        # execute view function
        response = endpoint(context, request)
        # try to validate response body
        if request.response.content_type == 'application/json':
            interprete_validation_instructions(json.loads(response.text),
                                               validations, 'response')
        return response
    return validate

validate_json_schema.options = ('is_error_view',)


def interprete_validation_instructions(json_data, validations, req_res):
    """
    Run JSON schema validations against JSON data.

    The validations list consists of dictionaries with the following spec:
      - schema: my-schema.json   [required]
        apply-to: some-key       [optional, applies to all data of not given
        in: response             [required, one of (request|response)]
        repeat: True             [optional, defaults to False]

    Schema files are expected to reside in the folder 'docs/docson/schemas'.
    The policy is to raise Exceptions when the validation instructions are
    malformed, but only log warnings when validation cannot happen.
    """
    if not validations:
        return
    logger = get_logger(__name__)
    if not isinstance(validations, list):
        raise BadValidationInstructions(_(
            'validations-is-not-list',
            default='Validations should be a list.'))
    else:
        for validation in validations:
            for required in ('schema', 'in'):
                if required not in validation:
                    raise BadValidationInstructions(_(
                        'bad-validation-instructions-missing-field',
                        default='Missing field: ${field}',
                        mapping={'field': required}))
            if validation['in'] not in ('request', 'response'):
                raise BadValidationInstructions(_(
                    'bad-validation-instructions-setting-in',
                    'Setting "in" should be either "request" or "response"'))
            if validation['in'] != req_res:
                continue
            data = json_data
            if 'apply-to' in validation:
                data = json_data[validation['apply-to']]
            if 'repeat' in validation and validation['repeat']:
                if not isinstance(data, list):  # dict too?
                    logger.warning('Validation for schema %s cannot'
                                   ' be repeated on %s (type %s).',
                                   validation['schema'],
                                   validation['apply-to'],
                                   type(data))
                else:
                    for dat in data:
                        apply_schema(dat, validation['schema'])
            else:
                apply_schema(data, validation['schema'])


def apply_schema(data, schema_name):
    """Find and apply a JSON schema to JSON data and deal with any problems."""
    # get schema location
    logger = get_logger(__name__)
    p2s = '{}/docs/docson/schemas'\
          .format('/'.join(os.path.abspath(__file__).split('/')[:-1]))
    schema_path = '{}/{}'.format(p2s, schema_name)
    if os.path.exists(schema_path):
        schema_str = open(schema_path, 'r').read()
        # redirect shared subschema pointers to our file structure
        our_file = '"file:{}/shared/'.format(p2s)
        schema_str = schema_str.replace('"./shared/', our_file)
        schema_str = schema_str.replace('"/shared/', our_file)
        schema = json.loads(schema_str)
        try:
            if not isinstance(data, dict):  # does not make sense
                logger.warning("The data %s cannot be validated"
                               "against a schema", data)
            jsonschema.validate(data, schema)
        except (KeyError, jsonschema.ValidationError) as val_exc:
            raise InvalidResponse(_('invalid-response',
                                    default=val_exc.message))
        except (AttributeError, jsonschema.SchemaError) as sch_exc:
            logger.error("The schema %s is not valid: %s.",
                         schema_path, str(sch_exc))
            raise BadValidationInstructions(_(
                'bad-validation-instructions-bad-schema',
                default="The schema ${schema} is not valid",
                mapping={'schema': schema_name}))
    else:
        logger.error("The schema %s could not be found.", schema_path)
        raise BadValidationInstructions(_(
            'bad-validation-instructions-schema-not-found',
            default="The required schema ${schema} could not be found to "
                    "perform validation.",
            mapping={'schema': schema_name}))



"""
Deal with content types in a generic way. Negotiate incoming content types.
"""
from mimetypes import guess_type

from spynl.main.serial import xml, json, py, html, yaml, csv
from spynl.main.serial.exceptions import UndeterminedContentTypeException
from spynl.main.serial.exceptions import UnsupportedContentTypeException


handlers = {'application/json': {'dump': json.dumps,
                                 'load': json.loads,
                                 'sniff': json.sniff},
            'application/xml': {'dump': xml.dumps,
                                'load': xml.loads,
                                'sniff': xml.sniff},
            'application/x-yaml': {'dump': yaml.dumps,
                                   'load': yaml.loads,
                                   'sniff': yaml.sniff},
            'text/csv': {'dump': csv.dumps,
                         'load': csv.loads,
                         'sniff': csv.sniff},
            'text/html': {'dump': html.dumps},
            'text/x-python': {'dump': py.dumps}}
handlers['text/xml'] = handlers['application/xml']


def negotiate_request_content_type(request):
    """
    Negotiate and return a content type for this request.
    Priorisation:
    1. Content-Type Header
    2. Body sniffing
    Throw an Exception when can't find a type (but have a body) or
    if we end up with a type we don't know
    (but the request method is POST|PUT).
    """
    content_type = request.content_type
    if not content_type or content_type not in handlers:
        for key, values in handlers.items():
            if 'sniff' in values and values['sniff'](request.text):
                return key

    if not content_type:
        if not request.body:
            return None
        raise UndeterminedContentTypeException()

    if content_type not in handlers and request.method in ('POST', 'PUT'):
        raise UnsupportedContentTypeException(content_type)

    return content_type


def negotiate_response_content_type(request):
    """
    Negotiate and return a content type to be used in the response.

    Priorisation:
    1. file extension in the URL
    2. Accept Header (only first mentioned type)
    3. Originally set Content-Type (e.g. browser default or set by a view)
    4. Spynl default (application/json)
    """
    content_type = None
    if hasattr(request, 'path_extension'):
        if request.path_extension:
            content_type = request.path_extension

    accept_type = None
    if 'Accept' in request.headers:
        accept_type = request.headers['Accept'].split(',')[0]
    # we use accept only if it is not one of the general types
    if not content_type and accept_type\
            and accept_type not in ('text/html', '*/*'):
        content_type = accept_type

    if not content_type:
        content_type = request.response.content_type

    if not content_type:
        content_type = 'application/json'

    # make proper type description if we only have a short name (e.g. 'csv')
    if content_type and '/' not in content_type:
        content_type = guess_type('file.' + content_type)[0]

    return content_type


def supported_types():
    """Return the types supported for serialisation"""
    return list(handlers.keys())



"""Module to find and include all spynl plugins"""
from pkg_resources import iter_entry_points  # pylint: disable=E0611
import logging


def include(config, name, _plugins=None):
    """
    Include a plugin. First, check if it is already included. Then, check its
    requirements (the plugin should implement the 'requires'-method for that).
    Finally, include the plugin (pyramid then calls the plugin's 'includeme'
    method, see http://pyramid.readthedocs.org/en/latest/api/config.html).
    """
    log = logging.getLogger(__name__)
    if _plugins is None:
        _plugins = config.get_settings()['plugins']

    if name not in _plugins:
        log.info('Initialising plugin %s', name.replace('.plugger', ''))
        _plugins.append(name)

        module = config.maybe_dotted(name)

        if hasattr(module, 'requires'):
            for name in module.requires():
                if not name.endswith('.plugger'):
                    name = name + '.plugger'
                include(config, name, _plugins)

        config.include(module)


def find(config):
    """
    Find all plugins.
    Either they are explicitly specified in settings or we find them in the
    'spynl'-namespace (which is declared in spynl.__init__). In that
    case, we look for entry_points in the group 'spynl.plugins', which they
    define in their setup.py.
    Finally, we call include on them.
    """
    log = logging.getLogger(__name__)
    settings = config.get_settings()
    plugins = settings['plugins']

    if 'enable_plugins' in settings:
        enable_plugins = settings['enable_plugins']
    else:
        enable_plugins = [e.name for e in
                          iter_entry_points('spynl.plugins')]

    for plugin in enable_plugins:
        log.debug('Found plugin %s', plugin.replace('.plugger', ''))
        include(config, plugin, plugins)


class PluginsLoaded(object):
    """ Event to signal that all plugins have been loaded"""
    def __init__(self, config):
        self.config = config


def main(config):
    """initialize this module, find all plugins and include them"""
    config.add_settings(plugins=[])

    config.add_directive('include_plugin', include)
    config.add_directive('find_plugins', find)
    config.find_plugins()



"""
Here we define more custom (de)serialisations for objects.
Here, we do text/unicode(utf-8) conversions.
Furthermore, we apply (de)serialisation which plugins can define for object
types like IDs or dates.
"""
from pyramid import threadlocal
from spynl.main.dateutils import (date_format_str, localize_date,
                                  date_to_str, date_from_str)
from spynl.main.utils import get_settings, get_logger
from spynl.main.locale import SpynlTranslationString as _


class SpynlDecoder(object):
    """
    (Incoming) Decodes Python objects from strings.
    It provides a function that routes to custom
    decoders for certain fields.
    """

    def __init__(self, context=None):
        """Enable the decoding functions to be context-aware."""
        self.context = context

    def __call__(self, dic):
        """
        Decode dic.
        In the setting serial_decode_functions, decoding functions are defined
        for specific fields. We use those functions for those fields.

        In the end, all keys that are still a string, are decoded as unicode.

        All the custom decode functions need to get the dic as an argument,
        so they change the dic and not a copy.
        They also get the context of the request.
        """
        settings = get_settings()
        decode_functions = settings.get('serial_decode_functions', {})
        for fieldname in dic:
            if fieldname in decode_functions:
                decode_functions[fieldname](dic, fieldname=fieldname,
                                            context=self.context)

        # All remaining strings: we use unicode internally
        # & we expect incoming strings to be UTF-8 encoded
        for k in dic:
            if isinstance(dic[k], bytes):
                dic[k] = str(dic[k], errors='strict', encoding='utf-8')

        return dic


def encode(obj):
    """
    (Outgoing) Encodes a Python object to str.

    In serial_encode_functions, functions are defined for specific object
    types. We use those specific functions to encode those types.
    """
    # type specific casting:
    settings = get_settings()
    encode_functions = settings.get('serial_encode_functions', {})
    for obj_type in encode_functions:
        if isinstance(obj, obj_type):
            obj = encode_functions[obj_type](obj)

    return str(obj)


def add_decode_function(config, function, fields):
    """
    Add the specified function to the fields in the serial_decode_functions
    dictionary.
    example serial_decode_functions: {'_id': decode_id, 'date': decode_date}
    """
    log = get_logger('spynl.main.serial')
    settings = config.get_settings()
    decode_functions = settings.get('serial_decode_functions', {})
    for field in fields:
        if decode_functions.get(field) is not None:
            log.warning('A custom decoding function for field %s is being '
                        'overwritten.', field)
        decode_functions[field] = function
    if function.__code__.co_argcount != 3:
        log.warning('Custom decoding functions should have three arguments.')
    # line below needed if setting was not initialised before
    config.add_settings(serial_decode_functions=decode_functions)


def add_encode_function(config, function, obj_type):
    """
    Add the specified function to the types in the serial_encode_functions
    dictionary.
    example serial_decode_functions: {bool: encode_boolean, datetime:
    encode_datetime}
    """
    log = get_logger('spynl.main.serial')
    settings = config.get_settings()
    encode_functions = settings.get('serial_encode_functions', {})
    if encode_functions.get(obj_type) is not None:
        log.warning('You are replacing the encoding function for type %s',
                    obj_type)
    encode_functions[obj_type] = function
    # line below needed if setting was not initialised before
    config.add_settings(serial_encode_functions=encode_functions)


def decode_date(dic, fieldname, context):
    """Localize date to UTC"""
    try:
        dic[fieldname] = localize_date(date_from_str(dic[fieldname]), tz='UTC')
    except Exception:
        raise ValueError(_(
            'date-decode-value-error',
            default="The value '${value}' for key '${key}' does not seem to "
                    "be a valid date string that conforms to ${format}.",
            mapping={'value': dic[fieldname], 'key': fieldname,
                     'format': date_format_str()}))


def encode_boolean(obj):
    """lower the case to get a json/js boolean"""
    return str(obj).lower()


def encode_date(obj):
    """localize the date and make it a string"""
    return date_to_str(localize_date(obj))


def encode_spynl_translation_string(obj):
    """ determine the locale of the request and translate the string """
    request = threadlocal.get_current_request()
    # import pdb; pdb.set_trace()
    localizer = request.localizer
    return obj.translate(localizer)



"""Generic custom exceptions for all packages to use."""

from pyramid.httpexceptions import HTTPBadRequest, HTTPForbidden

from spynl.main.locale import SpynlTranslationString as _


class SpynlException(Exception):
    """
    The Superclass for all Spynl-specific Exceptions.

    If your Exception inherits from this, the Spynl Error handling
    can treat it differently, e.g. show its message.
    You can also specify with which HTTP exception it should be
    escalated.
    """
    http_escalate_as = HTTPBadRequest

    def __init__(self, message='an internal error has occured'):
        self.message = message

    def __str__(self):
        """
        This will return a str version of the message. If the message is a
        SpynlTranslationString, it will return an interpolated version of the
        default (no translation). """
        return str(self.message)


class BadOrigin(SpynlException):
    """Bad origin exception."""

    http_escalate_as = HTTPForbidden

    def __init__(self, origin):
        """Set the origin attribute."""
        self.origin = origin
        self.message = _(u'bad-origin',
                         default="Requests to the Spynl API are not "
                         "permitted from origin '${origin}'.",
                         mapping={'origin': self.origin})


class IllegalAction(SpynlException):
    """Raise if the desired action is not allowed."""

    def __init__(self, message):
        """Exception message."""
        self.message = message


class MissingParameter(SpynlException):
    """Exception when parameter is missing."""

    def __init__(self, param):
        """Exception message."""
        self.message = _('missing-parameter',
                         default='Missing required parameter: ${param}',
                         mapping={'param': param})


class IllegalParameter(SpynlException):
    """Exception when parameter is illegal."""

    def __init__(self, param):
        """Exception message."""
        self.message = _('illegal-parameter',
                         default='Illegal parameter: ${param}',
                         mapping={'param': param})


class BadValidationInstructions(SpynlException):
    """Exception when the validation documentation cannot be used."""

    def __init__(self, error):
        """Exception message."""
        self.message = _('bad-validation-instructions',
                         default='The description of validations for this'
                         ' endpoint cannot be used: ${error}',
                         mapping={'error': error})


class InvalidResponse(SpynlException):
    """Exception when the response should be validated, but could not."""

    def __init__(self, error):
        """Exception message."""
        self.message = _('invalid-response',
                         default='Spynl could not generate a valid response: ${error}',
                         mapping={'error': error})


class EmailTemplateNotFound(SpynlException):
    """Exception when email template file is not found."""

    def __init__(self, template):
        """Exception message."""
        self.message = _('email-tmpl-not-found',
                         default='The email template <${template}> was not found.',
                         mapping={'template': template})



"""
The about plugin is used to display meta data about the application, e.g.
database connection or build information.
"""
from pyramid.security import NO_PERMISSION_REQUIRED
from pyramid.settings import asbool
from pyramid.security import Allow, Authenticated, DENY_ALL

from spynl.main.routing import Resource
from spynl.main.about.endpoints import (hello, versions, build, environment,
                                        endpoint_doc, schemas, ini)


class AboutResource(Resource):
    """The resource class for /about."""

    paths = ['about']

    __acl__ = [(Allow, 'role:sw-developer', 'read'),
               DENY_ALL]


class StaticResource(Resource):
    """The resource class for static assets. Open if you're authenticated."""

    __acl__ = [(Allow, Authenticated, 'read'), DENY_ALL]




def main(config):
    """doc, get, version, build, add endpoints."""
    settings = config.get_settings()

    config.add_static_view(name='static_swagger',
                           path='spynl.main:docs/swagger-ui/',
                           factory=StaticResource,
                           permission='read')
    config.add_static_view(name='static_docson',
                           path='spynl.main:docs/docson/',
                           permission='read')

    # check if we can use authentication for the more sensitive views
    permission = NO_PERMISSION_REQUIRED
    if asbool(settings.get('spynl.auth', True)):
        permission = 'read'
    else:
        permission = NO_PERMISSION_REQUIRED

    config.add_endpoint(hello, None, context=AboutResource,
                        permission=NO_PERMISSION_REQUIRED)
    config.add_endpoint(versions, 'versions', context=AboutResource,
                        permission=permission)
    config.add_endpoint(build, 'build', context=AboutResource,
                        permission=NO_PERMISSION_REQUIRED)
    config.add_endpoint(environment, 'environment', context=AboutResource,
                        permission=permission)
    config.add_endpoint(endpoint_doc, 'doc', context=AboutResource,
                        permission=Authenticated)  # TODO: deprecate
    config.add_endpoint(endpoint_doc, 'endpoints', context=AboutResource,
                        permission=Authenticated)
    config.add_endpoint(schemas, 'schemas', context=AboutResource,
                        permission=Authenticated)
    config.add_endpoint(ini, 'ini', context=AboutResource,
                        permission=permission)



"""Handle JSON content."""

import json
import re

from spynl.main.serial import objects
from spynl.main.serial.exceptions import MalformedRequestException


def loads(body, headers=None, context=None):
    """Return body as JSON."""
    try:
        decoder = objects.SpynlDecoder(context)
        return json.loads(body, object_hook=decoder)
    except ValueError as e:
        raise MalformedRequestException('application/json', str(e))


def dumps(body, pretty=False):
    """Return JSON body as string."""
    indent = None
    if pretty:
        indent = 4

    class JSONEncoder(json.JSONEncoder):
        def default(self, obj):  # pylint: disable=method-hidden
            return objects.encode(obj)
    return json.dumps(body, indent=indent, ensure_ascii=False, cls=JSONEncoder)


def sniff(body):
    """
    sniff to see if body is a json object.

    Body should start with any amount of whitespace and a {.
    """
    expression = re.compile(r'^\s*\{')
    return bool(re.match(expression, body))



"""
The serial module handles incoming and outgoing data streams.
It performs serialisation/deserialisation of the whole stream (for several data
formats,e.g. json, xml), as well as  some for specific data (e.g. for dates).

Our policy is as follows:
* We deal with text internally as unicode
* We expect incoming strings to be UTF-8 encoded.
* As internal representation of incoming data we use a dictionary, with values
either being text or specific objects (e.g. datetime objects).
* Everything outgoing is serialised to a unicode text, as well as the data
format allows (see e.g. the .csv part).
Here, the module can also do some pretty-printing if wanted.

The main utility functions of this module are parse_post_data and renderer,
they are added to config in main.
"""

from datetime import datetime

from pyramid.httpexceptions import HTTPBadRequest
from pyramid.settings import asbool

from spynl.main.serial.typing import handlers
from spynl.main.serial.typing import negotiate_response_content_type
from spynl.main.serial.exceptions import (UndeterminedContentTypeException,
                                          UnsupportedContentTypeException,
                                          SerializationUnsupportedException,
                                          DeserializationUnsupportedException,
                                          MalformedRequestException)
from spynl.main.serial.objects import (add_decode_function, decode_date,
                                       add_encode_function, encode_boolean,
                                       encode_date,
                                       encode_spynl_translation_string)
from spynl.main.locale import SpynlTranslationString


def parse_post_data(request):
    """Parse data according to the requests content type."""
    try:
        context = hasattr(request, 'context') and request.context or None
        parsed_body = loads(request.text, request.content_type,
                            request.headers, context)
    except MalformedRequestException as e:
        raise HTTPBadRequest(detail=e.message.translate(request.localizer))

    return parsed_body


def renderer(values, system):
    """
    Render data which the view created as a response.

    Spynl endpoints usually return dictionaries as responses, which
    this module renders in the content type set on the request
    (see parse_post_data).
    If spynl.pretty is set, formatting (pretty-printing) is done.
    """
    r = system['request']
    if not isinstance(system['context'], Exception):
        r.response.content_type = negotiate_response_content_type(r)

    if r.response.content_type == 'application/json'\
       and 'status' not in values:
        values['status'] = 'ok'

    pretty = asbool(r.registry.settings.get('spynl.pretty'))
    try:
        response = dumps(values, r.response.content_type, pretty=pretty)
    except UnsupportedContentTypeException as e:
        raise UndeterminedContentTypeException(str(e))

    return response


# ---- high-level dumps and loads functions,
#      relay to specific dumps and loads per type

def dumps(body, content_type, pretty=False):
    """Relay to content-type specific dumping."""
    try:
        handler = handlers[content_type]
    except KeyError:
        raise UnsupportedContentTypeException(content_type)

    try:
        dump = handler['dump']
    except KeyError:
        raise SerializationUnsupportedException(content_type)

    return dump(body, pretty=pretty)


def loads(body, content_type, headers=None, context=None):
    """Relay to content-type specific load."""
    if not body:
        return {}

    try:
        handler = handlers[content_type]
    except KeyError:
        raise UnsupportedContentTypeException(content_type)

    try:
        load = handler['load']
    except AttributeError:
        raise DeserializationUnsupportedException(content_type)

    return load(body, headers, context)


def main(config):
    """
    Add our main utility functions to be used in the config,
    to render output and parse post data
    Define decoding and encoding functions for objects
    """
    config.add_settings({'spynl.renderer': renderer,
                         'spynl.post_parser': parse_post_data})
    # define decode function for date fields:
    add_decode_function(config, decode_date, ['date'])
    # define encoding functions
    add_encode_function(config, encode_date, datetime)
    add_encode_function(config, encode_boolean, bool)
    add_encode_function(config, encode_spynl_translation_string,
                        SpynlTranslationString)



import os
from invoke import task
from invoke.exceptions import Exit
import pip

from spynl.main.version import __version__ as spynl_version
from spynl.main.pkg_utils import  get_config_package, get_dev_config 
from spynl.cli.utils import resolve_package_names, package_dir, chdir


packages_help = "Affected packages, defaults to all installed."


@task(help={'scm-url': 'The URL needed to clone the repo, either for '
                       'git or mercurial. Please no revision information, '
                       'you can use the revision parameter for that',
            'developing': 'If True (default), python setup.py develop is '
                          'used, otherwise install',
            'revision': 'Which revision to update to (default: do not update)',
            'fallbackrevision': 'Use this revision if the repository does not '
                                ' have the target revision.',
            'install-deps': 'If True (default), pre-install dependencies '
                            '(eg. via apt-get).',
            'src-path': 'Desired installation path. Defaults to '
                        '$VIRTUAL_ENV/src directory.'})
def install(ctx, scm_url=None, developing=True, revision=None,
            fallbackrevision=None, install_deps=True, src_path=None):
    '''
    Clone a repo, update to revision or fallbackrevision
    and install the Spynl package.
    '''
    ctx.run('pip install --upgrade setuptools')  # make sure we have the latest
    src_path = src_path or os.environ['VIRTUAL_ENV'] + '/src'
    if scm_url is None:
        raise Exit("[spynl dev.install] Please give the scm_url parameter.")
    scm_type = (scm_url.endswith('.git') or '.git@' in scm_url) and "git" or "hg"
    if fallbackrevision is None:
        fallbackrevision = scm_type == 'git' and 'master' or 'default' 
    if not os.path.exists(src_path):
        os.mkdir(src_path)
    repo_name = scm_url.strip('/')[-1]
    if scm_type == 'git' and repo_name.strip('/').endswith('.git'):
        repo_name = repo_name.split('/')[:-4]
    repo_path = src_path + '/' + repo_name
    if not os.path.exists(repo_path):
        with chdir(src_path):
            print("[spynl dev.install] CLONING from %s" % scm_url)
            ctx.run('%s clone %s %s' % (scm_type, scm_url, repo_name))
    else:
        print("[spynl dev.install] UPDATING {}".format(repo_name))
        with chdir(repo_path):
            cmd = scm_type == 'git' and 'git pull' or 'hg pull --update'
            ctx.run(cmd)
    with chdir(repo_path):
        if revision is not None:
            cmd = scm_type == 'git' and 'git checkout' or 'hg update'
            print('[spynl dev.install] Using revision %s ...' % revision)
            if not ctx.run('{} {}'.format(cmd, revision), warn=True):
                print('[spynl dev.install] Updating to fallback revision %s ...'
                        % fallbackrevision)
                ctx.run('{} {}'.format(cmd, fallbackrevision))
        if os.path.isfile('./setup.sh'):
            pre_cmd = './setup.sh --pre-install '\
                        '--virtualenv=%s' % os.environ['VIRTUAL_ENV']
            if install_deps:
                pre_cmd += ' --install-dependencies'
            ctx.run(pre_cmd)
        if developing:
            print("[spynl dev.install] DEVELOPING {}".format(repo_name))
            ctx.run('python setup.py develop')
        else:
            print("[spynl dev.install] INSTALLING {}".format(repo_name))
            ctx.run('python setup.py install')
        if os.path.isfile('./setup.sh'):
            ctx.run('./setup.sh --post-install')


@task
def serve(ctx):
    '''Run local dev server'''
    config_package = get_config_package()
    if config_package is None:
        raise Exit("No spynl-plugin found with .ini files. Exiting ...")
    with package_dir('spynl'):
        ctx.run('pserve %s/development.ini --reload'
                % config_package.location, pty=True)


@task(aliases=('tests',),
      help={'packages': packages_help,
            'called-standalone': 'If False, this task is a pre-task and the '
                                 'user gets to cancel the task flow when '
                                 'tests fail.',
            'reports': 'If True, Junit and Coverage reports will be made '
                       '(requires a few extra packages, '
                       'e.g. for pycoverage).'})
def test(ctx, packages='_all', called_standalone=True, reports=False):
    '''
    Perfom tests in one or more spynl plugins.
    '''
    for package_name in resolve_package_names(packages):
        print("[spynl dev.test] Testing repo: {}".format(repo_name))
        with package_dir(package_name):
            if not reports:
                result = ctx.run('py.test', warn=True)
            else:
                result = ctx.run('py.test --junit-xml=pytests.xml --cov %s '
                                 ' --cov-report xml --cov-append' % package_name)
            if not result.ok and called_standalone is False\
               and input("[spynl dev.test] Tests failed. Continue anyway? Y/n") not in ('y', 'Y'):
                raise Exit("[spynl dev.test] Aborting at user request.")


@task(help={'packages': packages_help,
            'languages': 'An iterable of language codes. Defaults to ("nl",).',
            'action': 'Either "compile" (compile selected languages) or '
                      '"refresh" (extract messages from the source code '
                      'and update the .po files for the selected languages). '
                      'Defaults to compile.'})
def translations(ctx, packages='_all', languages=('nl',), action='compile'):
    '''
    Ensure that translations files are up-to-date w.r.t. to the code. If action
    is set to compile (default), this will compile the catalogs for all selected
    languages. If action is set to refresh, this will extract messages from the
    source code and update the .po files for the selected languages (will
    initialize if necessary).
    '''
    if action == 'compile':
        refresh = False
    elif action == 'refresh':
        refresh = True
    else:
        raise Exit("[spynl dev.translations] action should be 'compile' or "
                   "'refresh'.")

    for package_name in resolve_package_names(packages):
        with package_dir(package_name):
            if package_name == 'spynl':
                package_name = 'spynl.main'
            print("[spynl dev.translations] Package: %s ..." % package_name)
            package_path = package_name.replace('.', '/')
            # make locale folder if it doesn't exist already:
            if not os.path.exists('%s/locale' % package_path):
                os.mkdir('./%s/locale' % package_path)
            if refresh:
                # Extract messages from source:
                ctx.run('python setup.py extract_messages '
                        '--output-file {rp}/locale/messages.pot --no-wrap '
                        '--sort-by-file --input-dirs {rp} --project {project} '
                        '--copyright-holder "Softwear BV" --version {v}'
                        .format(rp=package_path, project=package_name,
                                v=spynl_version))
            for lang in languages:
                path2po = '%s/locale/%s/LC_MESSAGES/%s.po' % (package_path, lang,
                                                              package_name)
                # update if refresh
                if refresh:
                    # init if needed:
                    if not os.path.exists(path2po):
                        print('[spynl dev.translations] File %s does not exist.'
                              ' Initializing.' % path2po)
                        ctx.run('python setup.py init_catalog -l {lang} '
                                '-i {rp}/locale/messages.pot '
                                '-d {rp}/locale -D {rn}'
                                .format(rp=package_path, rn=package_name, lang=lang))
                    # update if not init
                    else:
                        print('[spynl dev.translations] update the %s catalog'
                              % lang)
                        ctx.run('python setup.py update_catalog -N --no-wrap '
                                '-l {lang} -i {rp}/locale/messages.pot '
                                '-d {rp}/locale -D {rn}'
                                .format(rp=package_path, rn=package_name, lang=lang))
                # if not refresh, compile
                elif os.path.exists(path2po):
                    ctx.run('python setup.py compile_catalog --domain {rn} '
                            '--directory {rp}/locale --domain {rn} '
                            '--locale {lang}'
                            .format(rp=package_path, rn=package_name, lang=lang))
                else:
                    print('[spynl dev.translations] File %s does not exist.'
                          ' Update the package first.' % path2po)

                print("[spynl dev.translations] Done with language %s." % lang)
                print("--------------------------------------------------")



"""The main package of Spynl."""


import os.path

from pyramid.config import Configurator
from pyramid.settings import asbool
from pyramid.security import NO_PERMISSION_REQUIRED
from pyramid.viewderivers import INGRESS

from spynl.main import (plugins, serial, about,
                        routing, events, endpoints, session)
from spynl.main.utils import (renderer_factory, check_origin,
                              handle_pre_flight_request)
from spynl.main.exceptions import SpynlException
from spynl.main.error_views import spynl_error, error400, error500
from spynl.main.validation import validate_json_schema

from spynl.main.docs.documentation import make_docs
from spynl.main.docs.settings import check_required_settings
from spynl.main.dateutils import now
from spynl.main.locale import TemplateTranslations



def main(global_config, test_settings=None, **settings):
    """
    Return a Pyramid WSGI application.

    Before that, we tell plugins how to add a view and tell views which
    renderer to use. And we take care of test settings. Then, we initialise the
    main plugins and the external plugins (which are not in this repository).
    """
    # load (test) settings
    config = Configurator(settings=settings)
    config.add_settings({'spynl.start_time': now()})

    if test_settings:
        for key, value in test_settings.items():
            config.add_settings({key: value})

    # Add spynl.main's view derivers
    config.add_view_deriver(validate_json_schema)
    config.add_view_deriver(handle_pre_flight_request,
                            under=INGRESS) # do this first
    config.add_view_deriver(check_origin)

    # initialize the main plugins
    # serial should be before plugins, because plugins can overwrite field
    # treatment
    # session needs to be after plugins, because plugins can set the session
    # mechanism
    routing.main(config)
    events.main(config)
    serial.main(config)
    endpoints.main(config)
    about.main(config)
    plugins.main(config)
    session.main(config)

    check_required_settings(config)

    # Custom renderer from main.serial or vanilla json renderer
    config.add_renderer('spynls-renderer', renderer_factory)

    config.add_translation_dirs('spynl.main:locale/')

    # Error views
    config.add_view(error400, context='pyramid.httpexceptions.HTTPError',
                    renderer='spynls-renderer',
                    permission=NO_PERMISSION_REQUIRED,
                    is_error_view=True)
    config.add_view(spynl_error, context=SpynlException,
                    renderer='spynls-renderer',
                    permission=NO_PERMISSION_REQUIRED,
                    is_error_view=True)
    config.add_view(error500, context=Exception,
                    renderer='spynls-renderer',
                    permission=NO_PERMISSION_REQUIRED,
                    is_error_view=True)

    # make spynl documentation
    make_docs(config)

    # add jinja for templating
    config.include('pyramid_jinja2')
    config.add_settings({
        'jinja2.filters': {'static_url': 'pyramid_jinja2.filters:static_url_filter'},
        'jinja2.i18n.gettext': TemplateTranslations})

    # initialise mailer, use dummy mailer for tests
    if test_settings:
        config.include('pyramid_mailer.testing')
    else:
        config.include('pyramid_mailer')

    config.commit()
    config.registry.notify(plugins.PluginsLoaded(config))


    return config.make_wsgi_app()



"""Sets up Spynl."""


import os
import sys
import codecs
import re
from setuptools import setup


here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.md')).read()


install_requires = ['webob==1.6.3', 'pyramid>=1.7,<1.8', 'invoke>=0.14,<0.15', 'requests',
                    'pyramid_mailer', 'pyramid_jinja2', 'pyramid_exclog',
                    'pytz', 'pbkdf2', 'python-dateutil', 'html2text',
                    'beaker', 'waitress', 'gunicorn', 'jsonschema',
                    'sphinx', 'pyyaml', 'tld', 'babel',
                    'webtest', 'pytest', 'pytest-raisesregexp']


def find_version(*file_paths):
    """ get the version string from a file path."""
    version_file = codecs.open(os.path.join(here, *file_paths), 'r').read()
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


def pytest_runner_dependency():
    """Return a list that contains the pytest_runner dependency if needed"""
    needs_pytest = {'pytest', 'test', 'ptr'}.intersection(sys.argv)
    return ['pytest-runner'] if needs_pytest else []


setup(name='spynl',
      version=find_version('spynl', 'main', 'version.py'),
      description='spynl',
      long_description=README,
      classifiers=["Programming Language :: Python :: 3.5",
                   "Framework :: Pylons",
                   "Topic :: Internet :: WWW/HTTP",
                   "Topic :: Internet :: WWW/HTTP :: WSGI :: Application"],
      author='Softwear BV',
      author_email='development@softwear.nl',
      url='http://www.softwear.nl',
      license='MIT',
      keywords='API SaaS',
      packages=['spynl.cli', 'spynl.main'],
      include_package_data=True,
      zip_safe=False,
      install_requires=install_requires,
      setup_requires=pytest_runner_dependency() + ['PasteScript'],
      test_suite="spynl",
      entry_points={
          "paste.app_factory": ["main = spynl.main:main"],
          "console_scripts": ["spynl = spynl.cli:program.run"]
      },
      paster_plugins=['pyramid']
)



"""
Utils to get info about Spynl packages
"""

import os
import configparser
from collections import namedtuple
import subprocess

from spynl.main.exceptions import SpynlException


def read_setup_py(path):
    """return contents of setup.py"""
    content = ''
    with open('%s/setup.py' % path, 'r') as setup:
        return ''.join(setup.readlines())


def get_spynl_packages():
    """
    Return the (locally) installed spynl-plugin packages.
    The term 'package' is the preferred term, synonymous with 'distribution':
    https://packaging.python.org/glossary/#term-distribution-package
    We ask for packages in a separate pip process, so we catch latest
    ones installed earlier in this context.
    """
    # mocking some package info pip.get_distributed_packages provides
    Package = namedtuple('Package', ['project_name', 'version', 'location'])
    installed_packages = []
    # --format json does not return locations :(
    plist = subprocess.run('pip list -l --format columns',
                           shell=True, stdout=subprocess.PIPE,
                           universal_newlines=True)
    for pinfo in [line for line in plist.stdout.split("\n")[2:]
                  if line.strip() != ""]:
        pinfo = pinfo.split()
        pname = pinfo[0]
        if len(pinfo) > 2:  # we need a location
            pversion = pinfo[1]
            plocation = pinfo[2]
            package = Package(pname, pversion, plocation)
            if pname == 'spynl' or 'spynl.plugins' in read_setup_py(plocation):
                installed_packages.append(package)
    return installed_packages


def get_spynl_package(name, packages=None):
    """Return the installed spynl package."""
    if packages is None:
        packages = get_spynl_packages()
    return next(filter(lambda p: p.project_name == name, packages),
                None)


def get_config_package():
    """
    Return the config package. Complain if several packages have the 
    .ini files.
    """
    config_package = None
    packages = get_spynl_packages()
    for package in packages:
        plisting = os.listdir(package.location)
        if 'development.ini' in plisting and 'production.ini' in plisting:
            if config_package is not None:
                emsg = ("Two packages have configurations (development.ini "
                        "and production.ini): %s and %s. unsure which to use!"
                        % (config_package.project_name, package.project_name))
                raise SpynlException(emsg)
            config_package = package
    return config_package


def get_dev_config():
    """Return the ConfigParser for development.ini"""
    config_package = get_config_package()
    config = configparser.ConfigParser()
    config.read('%s/development.ini'  % config_package.location)
    return config['app:main']





"""
This module is used to parse JSON-like objects from the URL into a python
structure of dicts and lists.

Be aware that all values will be strings, so parse anything that you expect
to be something else, e.g. an int.

>>> loads_dict(request.GET)
"""

import re

from spynl.main.locale import SpynlTranslationString as _


WS = re.compile(r'\s+')
KEY = re.compile(r'([^:]*)')
LITERALS = {'object': re.compile(r'([^,\}]*)'),
            'array': re.compile(r'([^,\]]*)')}


def loads_dict(dictionary):
    """
    Read in a dict, parse each value into a python structure
    """
    result = []
    for k, v in dictionary.items():
        try:
            v = loads(v)
        except UnexpectedEndOfInput as e:
            extra = ' when processing argument "{}".'.format(k)
            e.args = (e.args[0][:-1] + extra,) + e.args[1:]
            raise
        result.append((k, v))
    return dict(result)


def loads(urlson):
    """
    This method runs the parser on any string and returns its collected
    end-product. We call the string we work on "urlson" throughout.
    """
    parser = Parser()
    value_ctx(urlson, parser)
    return parser.value


class Parser:
    """
    This parser uses a stack of lists and dicts to model the object structure
    we find. The functions value_ctx, array_ctx and object_ctx call this object
    whenever they ancounter a meaningful event.
    """

    def __init__(self):
        """initialise parser"""
        self.stack = [[{}, 'value']]

    def __call__(self, event, token):
        """call parser"""
        stack = self.stack
        top = stack[-1]
        value = top[0]

        if event == 'object_start':
            stack.append([{}, None])
        elif event == 'array_start':
            stack.append([[], None])
        elif event == 'key':
            top[1] = token
        elif event == 'literal':
            # decide if we have to append to a list or set in a dict
            if isinstance(value, list):
                value.append(token)  # pylint: disable=E1101
            elif isinstance(value, dict):  # should be one line (first test)
                dict_key = top[1]
                value[dict_key] = token
        elif event in ['object_stop', 'array_stop']:
            parent_value = stack[-2][0]
            if isinstance(parent_value, list):
                parent_value.append(value)
            elif isinstance(parent_value, dict):
                parent_key = stack[-2][1]
                parent_value[parent_key] = value

            stack.pop()

    @property
    def value(self):
        """return value"""
        return self.stack[0][0]['value']


class InvalidToken(ValueError):
    """Exception if token was invalid"""

    def __init__(self, urlson, offset):
        """initialise Token Error"""
        super().__init__()
        self.urlson = urlson
        self.offset = offset

    def __str__(self):
        """str representation of this Error"""
        context = self.urlson[self.offset:self.offset + 10]
        return 'Invalid token at offset ${offset}: ${context}...'\
                .format({'offset': self.offset, 'context': context})


class UnexpectedEndOfInput(ValueError):
    """Exeption if end of input was reached"""

    def __init__(self):
        """initialize"""
        super().__init__(
            _('unexpected-end-of-input', default='Unexpected end of input.'))


def handle_ws(urlson, offset=0):
    """return index of next non-whitespace character"""
    result = re.match(WS, urlson[offset:])
    if result:
        offset += result.end()
    return offset


def value_ctx(urlson, cb, offset=0, ctx=None):
    """
    This function parses a value of the GET string in.
    It tells the parser about literals or, alternatively, if objects or arrays
    are detected, it calls the appropriate parsing methods (array_ctx,
    object_ctx), who tell the parser what they found (they, in turn, do the
    same and also call value_ctx if they find something new which needs to be
    defined here).
    """
    if not ctx:
        ctx = ['root']

    ctx.append('value')

    offset = handle_ws(urlson, offset)

    if offset == len(urlson):
        cb('literal', '')
    elif urlson[offset] == '{':
        offset = object_ctx(urlson, cb, offset, ctx)
    elif urlson[offset] == '[':
        offset = array_ctx(urlson, cb, offset, ctx)
    elif ctx[-2] == 'root':
        cb('literal', urlson[offset:].strip())
        offset = len(urlson)
    elif ctx[-2] in ['array', 'object']:
        expr = LITERALS[ctx[-2]]
        result = re.match(expr, urlson[offset:])
        if result:
            cb('literal', result.groups()[0].strip())
            offset += result.end()
        else:
            cb('literal', '')
    else:
        raise InvalidToken(urlson, offset)

    ctx.pop()

    return offset


def array_ctx(urlson, cb, offset=0, ctx=None):
    """parse an Array"""
    if not ctx:
        ctx = ['root']

    ctx.append('array')
    offset += 1  # skip [

    cb('array_start', None)

    while 1:
        offset = handle_ws(urlson, offset)

        if offset == len(urlson):
            raise UnexpectedEndOfInput()
        if urlson[offset] == ']':
            offset += 1
            break
        elif urlson[offset] == ',':
            offset += 1
        else:
            offset = value_ctx(urlson, cb, offset, ctx)

    cb('array_stop', None)

    ctx.pop()

    return offset


def object_ctx(urlson, cb, offset=0, ctx=None):
    """parse an object"""
    if not ctx:
        ctx = ['root']

    ctx.append('object')
    offset += 1  # skip {

    cb('object_start', None)

    while 1:
        offset = handle_ws(urlson, offset)

        if offset == len(urlson):
            raise UnexpectedEndOfInput()
        elif urlson[offset] == '}':
            offset += 1
            break
        elif urlson[offset] == ',':
            offset += 1
        else:
            result = re.match(KEY, urlson[offset:])
            if result:
                offset += result.end()
                cb('key', result.groups()[0].strip())

                if offset == len(urlson):
                    raise UnexpectedEndOfInput()

                offset += 1  # skip :
                offset = value_ctx(urlson, cb, offset, ctx)
            else:
                raise InvalidToken(urlson, offset)

    cb('object_stop', None)

    ctx.pop()

    return offset
